#include <math.h>
#include <limits.h>
#include <float.h>
#include <time.h>
#include <assert.h>
#include <stdio.h>
#include <stdlib.h>

/* include the user-visible definitions for ASL.h */
#include "ASL_user.h"

/* ========================================================================== */
/* === macros =============================================================== */
/* ========================================================================== */

#define PRIVATE static
#define PUBLIC
#define ASLZERO ((ASLFLOAT) 0)
#define ASLONE ((ASLFLOAT) 1)
#define ASLTWO ((ASLFLOAT) 2)

#define ASLMIN(a,b) (((a) < (b)) ? (a) : (b))
#define ASLMAX(a,b) (((a) > (b)) ? (a) : (b))

#define ASLINF DBL_MAX
#define ASLINT_INF INT_MAX
#define ASLTRUE 1
#define ASLFALSE 0

/* turn off debugging (don't change this): */
#ifndef DEBUG
#define DEBUG
#endif

/* turn off printing (don't change this): */
#ifndef PRINT
#define PRINT
#endif

/* turn off printing of arrays and matrices (don't change this): */
#ifndef TESTPRINT
#define TESTPRINT
#endif

/* ========================================================================== */
/* Modify these lines only, to toggle debugging and/or printing */

/* uncomment this line to turn off debugging:*/
#undef DEBUG
/**/

/* uncomment this line to turn off PRINTF printing:*/
/*#undef PRINT*/
/**/

/* uncomment this line to turn off TESTPRINT printing:*/
/*#undef TESTPRINT*/
/**/

/* note that #include "mex.h" and #include <assert.h> must appear AFTER
   the NDEBUG definition is finalized */

/* ========================================================================== */

/* non-debug printf uses PRINTF */
#ifdef PRINT
#define PRINTF(e) printf e
#else
#define PRINTF(e)
#endif

/* debug printf uses DPRINT */
#ifdef DEBUG
#define DPRINT(e) printf e
#else
#define DPRINT(e)
#endif

#define ERROR(msg) \
{ \
    fprintf (stderr, "Error in %s, line %d\n", __FILE__, __LINE__) ; \
    error_handler (msg) ; \
}

typedef struct ASLcom_struct /* common variables */
{
    ASLobjective *user ; /* information passed to user when function or
                            gradient must be evaluated */
    ASLINT          nf ; /* total number of function evaluations */
    ASLINT          ng ; /* total number of gradient evaluations */
    ASLINT         nmu ; /* total number of steps in mu evaluation */
    ASLFLOAT         b ; /* right side */
    ASLFLOAT     norma ; /* infinity norm of a vector */
    ASLparm      *Parm ; /* parameter structure */
    void    (*value) (ASLobjective *) ; /* evaluate objective function */
    void     (*grad) (ASLobjective *) ; /* evaluate objective gradient */
    void  (*valgrad) (ASLobjective *) ; /* function & gradient if given*/
    /* if objective function is quadratic (Parm->QuadCost = TRUE) */
    void       (*Ad) (ASLobjective *) ; /* compute A*d if cost quadratic */
    ASLFLOAT    deriv2 ; /* d'Ad */
} ASLcom ;

/* prototypes */

PRIVATE void ASLevaluate
(
    ASLcom *Com,  /* ASLcom structure */
    char  *what   /* f = function, g = gradient, fg = function and gradient */
) ;

PRIVATE int ASLd
(
    ASLFLOAT       *d, /* search dirction */
    ASLFLOAT      *mu, /* input guess for mu, return final mu */
    ASLFLOAT    *xnew, /* xnew = x + d */
    ASLFLOAT     *Gtd, /* g'd */
    ASLFLOAT     *Err, /* KKT error */
    ASLFLOAT       *x, /* current iterate */
    ASLFLOAT       *g, /* gradient at x */
    ASLFLOAT       *a, /* linear constraint is a'x = b */
    ASLFLOAT       *b, /* linear constraint is a'x = b */
    ASLFLOAT      *lo, /* lower bounds */
    ASLFLOAT      *hi, /* upper bounds */
    int             n, /* dimension */
    ASLFLOAT lambda_k, /* constant in denominator of d */
    ASLcom       *Com  /* common variables */
) ;

PRIVATE void ASLval
(
    ASLFLOAT      *vC,
    ASLFLOAT      *dC,
    ASLFLOAT        C,
    ASLFLOAT lambda_k,
    ASLFLOAT       *x,
    ASLFLOAT       *g,
    ASLFLOAT       *a,
    ASLFLOAT      *lo,
    ASLFLOAT      *hi,
    int             n,
    ASLcom       *Com  /* common variables */
) ;

PRIVATE ASLFLOAT ASLbb
(
    ASLFLOAT alpha,
    int         mm, /* stop current cycle when Parm->MaxCycle - mm <= 0 */
    int          n, /* problem dimension */
    ASLcom    *Com  /* common variables */
) ;

PRIVATE void ASLgather
(
    ASLFLOAT *y, /* compressed vector */
    ASLFLOAT *x, /* current x */
    int      *i, /* indices associated with subspace */
    int       n  /* number of elements to extract from x */
) ;

PRIVATE void ASLscatter
(
    ASLFLOAT *x, /* scatter y into x */
    ASLFLOAT *y, /* compressed vector */
    int      *i, /* indices associated with subspace */
    int       n  /* number of elements to extract from x */
) ;

PRIVATE void ASLstep
(
    ASLFLOAT *xnew, /* updated x vector */
    ASLFLOAT    *x, /* current x */
    ASLFLOAT    *d, /* search direction */
    ASLFLOAT    st, /* stepsize */
    ASLINT       n  /* dimension */
) ;

PRIVATE void ASLstepi
(
    ASLFLOAT *xnew, /* updated x vector */
    int         *i, /* indices of xnew that change */
    ASLFLOAT    *x, /* current x */
    ASLFLOAT    *d, /* search direction */
    ASLFLOAT    st, /* stepsize */
    int         nz  /* number of components to be updated */
) ;

PRIVATE ASLFLOAT ASLdot
(
    ASLFLOAT *x, /* first vector */
    ASLFLOAT *y, /* second vector */
    ASLINT    n  /* length of vectors */
) ;

PRIVATE ASLFLOAT ASLdoti
(
    ASLFLOAT *x, /* first vector */
    int      *i, /* indices associated with elements of x */
    ASLFLOAT *y, /* second vector */
    int       n  /* length of x */
) ;

double ASLadd
(
    ASLFLOAT *x, /* vector */
    int       n  /* length of vector */
) ;

PRIVATE void ASLcopy
(
    ASLFLOAT *x, /* first vector */
    ASLFLOAT *y, /* second vector */
    ASLINT    n  /* length of vectors */
) ;

PRIVATE ASLFLOAT ASLmax
(
    ASLFLOAT *x,
    ASLINT    n
) ;

void ASLint_init
(
    int *x,  /* array to be initialized */
    int  s,  /* scalar */
    int  n   /* length of x */
) ;

PRIVATE void ASLpartialMinSort
(
    int      *i, /* indices of m largest numbers */
    ASLFLOAT *x, /* numbers to sort */
    int       m, /* find m smallest numbers */
    int       n  /* length of x */
) ;

PRIVATE void ASLpartialMaxSort
(
    int      *i, /* indices of m largest numbers */
    ASLFLOAT *x, /* numbers to sort */
    int       m, /* find m smallest numbers */
    int       n  /* length of x */
) ;

PRIVATE void ASLminHeapSort
(
    int      *i, /* indices with decreasing order based on x */
    int   *heap, /* min heap */
    ASLFLOAT *x, /* numbers to sort */
    int       n  /* size of heap */
) ;

PRIVATE void ASLmaxHeapSort
(
    int      *i, /* indices with increasing order based on x */
    int   *heap, /* max heap */
    ASLFLOAT *x, /* numbers to sort */
    int       n  /* size of heap */
) ;

PRIVATE void ASLminheap_build
(
    int   *heap, /* on input, an unsorted set of indices */
    ASLFLOAT *x, /* numbers to sort */
    int       n  /* number of elements to build into the heap */
) ;

PRIVATE int ASLminheap_delete  /* return new size of heap */
(
    int   *heap, /* containing indices into x, 1..n on input */
    ASLFLOAT *x, /* not modified */
    int       n  /* number of items in heap */
) ;

PRIVATE int ASLminheap_add
(
    int    leaf, /* the new leaf */
    int   *heap, /* size n, containing indices into x */
    ASLFLOAT *x, /* not modified */
    int       n  /* number of elements in heap not counting new one */
) ;

PRIVATE void ASLminheapify
(
    int   *heap, /* size n, containing indices into x */
    ASLFLOAT *x, /* not modified */
    int       p, /* start at node p in the heap */
    int       n  /* heap [1 ... n] is in use */
) ;

PRIVATE void ASLmaxheap_build
(
    int   *heap, /* on input, an unsorted set of elements */
    ASLFLOAT *x,
    int       n  /* number of elements to build into the heap */
) ;

PRIVATE int ASLmaxheap_delete  /* return new size of heap */
(
    int   *heap, /* containing indices into x, 1..n on input */
    ASLFLOAT *x, /* not modified */
    int       n  /* number of items in heap */
) ;

PRIVATE int ASLmaxheap_add
(
    int   *heap, /* size n, containing indices into x */
    ASLFLOAT *x, /* not modified */
    int    leaf, /* the new leaf */
    int       n  /* number of elements in heap not counting new one */
) ;

PRIVATE void ASLmaxheapify
(
    int   *heap, /* size n, containing indices into x */
    ASLFLOAT *x, /* not modified */
    int       p, /* start at node p in the heap */
    int       n  /* heap [1 ... n] is in use */
) ;

PRIVATE ASLFLOAT ASLnapsack /* return the final lambda */
(
    ASLFLOAT     *x, /* holds y on input, and the solution x on output */
    ASLFLOAT lambda, /* starting guess for lambda */
    ASLFLOAT     *a, /* a'x = b (linear constraint vector) */
    ASLFLOAT      b, /* right side */
    ASLFLOAT    *lo, /* lower bounds */
    ASLFLOAT    *hi, /* upper bounds */
    int           n, /* problem dimension */
    ASLFLOAT     *w, /* work array of size n */
    int        *iw1, /* integer work array of size n */
    int        *iw2, /* integer work array of size n */
    ASLcom     *Com  /* common variables */
) ;

PRIVATE ASLFLOAT ASLnapup /* return lambda */
(
    ASLFLOAT        *y, /* input is target and output is solution */
    ASLFLOAT    lambda, /* initial guess for multiplier */
    ASLFLOAT        *a, /* a'x = b (linear constraint vector) */
    ASLFLOAT         b, /* right side */
    ASLFLOAT       *lo, /* lower bounds */
    ASLFLOAT       *hi, /* upper bounds */
    int              n, /* problem dimension */
    ASLFLOAT       atx, /* a'x */
    ASLFLOAT     a2sum, /* sum_j a_j^2 */
    ASLFLOAT *breakpts, /* break points */
    int    *bound_heap, /* work array */
    int     *free_heap, /* work array */
    ASLcom        *Com  /* common variables */

) ;

PRIVATE ASLFLOAT ASLnapdown /* return lambda */
(
    ASLFLOAT        *y,
    ASLFLOAT    lambda, /* initial guess for multiplier */
    ASLFLOAT        *a, /* a'x = b (linear constraint vector) */
    ASLFLOAT         b, /* right side */
    ASLFLOAT       *lo, /* lower bounds */
    ASLFLOAT       *hi, /* upper bounds */
    int              n, /* problem dimension */
    ASLFLOAT       atx, /* a'x */
    ASLFLOAT     a2sum, /* sum_j a_j^2 */
    ASLFLOAT *breakpts, /* break points */
    int    *bound_heap, /* work array */
    int     *free_heap, /* work array */
    ASLcom        *Com  /* common variables */
) ;

PRIVATE void ASLprint_parms
(
    ASLparm     *Parm, /* SSMparm structure to be printed */
    ASLFLOAT grad_tol, /* gradient tolerance */
    int             n  /* dimension */
) ;

PRIVATE void error_handler
(
    char *message
) ;
