#include <math.h>
#include <ctype.h>
#include <stdlib.h>
#include <stdio.h>
#include <time.h>
#include <sys/time.h>
#include <unistd.h>
#include <sys/resource.h>
#include "SVM_user.h"

#define FALSE 0
#define TRUE 1
#define MAX(a,b) (((a) > (b)) ? (a) : (b))
#define MIN(a,b) (((a) < (b)) ? (a) : (b))
#define ZERO ((double) 0)

typedef struct SVMproblem_struct /* passed to user's evaluation routines */
{
    float      *Ax ; /* numerical values in matrix */
    short      *Ai ; /* row indices */
    int        *Ap ; /* pointers to start of each column in Ax, Ai */
    short       *a ; /* label vector (+- 1) */
} SVMproblem ;

static double Timer ( void ) ;

void read_problem
(
    const char *filename, /* name of file containing problem */
    SVMproblem   *PB,
    int         *dim,
    int    *nfeature,
    short *Ax_is_1
) ;

int main (int argc, char **argv)
{
    short kernel, Ax_is_1, *a, *Ai ;
    int i, j, n, nfeature, nlo, nhi, nfree, status, *Ap ;
    float *Ax ;
    double C, gamma, tic, time1, time2, tol, atx, *x ;
    SVMstat Stat ;
    ASLparm ASLParm ;
    SVMparm Parm ;
    SVMproblem PB ;

    C  = atof (argv [3]) ;
    printf ("\n\nproblem: %s // kernel: %s // C: %e\n",
             argv [1]+9, argv [2], C) ;
    tic = Timer () ;
    read_problem (argv [1], &PB, &n, &nfeature, &Ax_is_1) ;
    a  = PB.a ;
    Ax = PB.Ax ;
    Ai = PB.Ai ;
    Ap = PB.Ap ;
    
    time1 = Timer () - tic ;
    printf ("time to read data: %e\n", time1) ;
    x = (double *) malloc (n*sizeof (double)) ;
    printf ("number of features: %i\n", nfeature) ;
    printf ("nonzeros in data: %i\n", Ap [n]) ;
    printf ("problem dimension: %i\n", n) ;
    printf ("matrix to be created in format %s\n", argv [2]) ;
    fflush (stdout) ;
    kernel = 0 ;
    gamma = 0 ;
    if ( strcmp (argv [2],"linear") ) /* if not linear, compute A */
    {
        if      ( !strcmp (argv [2], "poly") ) kernel = 1 ;
        else if ( !strcmp (argv [2], "rbf") )  kernel = 2 ;
        else if ( !strcmp (argv [2], "sig") )  kernel = 3 ;
        else
        {
            printf ("%s not a valid kernel\n", argv [2]) ;
            exit (0) ;
        }
    }

    /* parameters for SVM */
    tol = 1.e-3 ;
    SVMdefault (&Parm, nfeature) ;
    Parm.Ax_is_1 = Ax_is_1 ;
    Parm.PrintLevel = 0 ;
    Parm.serr1 = 1. ;
    if ( kernel == 0 ) Parm.nsub = 250 ;
 
    /* parameters for ASL */
    ASLdefault (&ASLParm) ;
    ASLParm.PrintLevel = 0 ;
    ASLParm.PrintFinal = FALSE ;
    ASLParm.lo_cut = 1.e-12 ; /* constraint active if within lo_cut */
    ASLParm.hi_cut = 1.e-12 ;
    ASLParm.QuadCost = TRUE ; /* the cost function is quadratic */
    ASLParm.a_is_1 = TRUE ;
    ASLParm.QuadLazy = FALSE ;
    ASLParm.bnd_is_scalar = TRUE ;
    ASLParm.lambda0Factor = 1.e-6 ;
    ASLParm.Subspace = FALSE ;
    ASLParm.PrintParms = FALSE ;
    ASLParm.epsmu0 = 1.e-11 ;  /* absolute error in mu computation */
    ASLParm.epsmu1 = 1.e-12 ;  /* relative error in mu computation */

    tic = Timer () ;
    status = SVMsubspace (x, Ax, Ai, Ap, a, C, n, nfeature, kernel, tol,
                 &Stat, &Parm, &ASLParm) ;
    time2 = Timer () - tic ;
    printf ("\n    =========================================\n") ;
    printf ("Status: %17i\n", status) ;
    printf ("Function value:     %20.10e\n", Stat.f) ;
    printf ("KKT error:          %20.10e\n", Stat.err) ;
    nfree = 0 ;
    nhi = 0 ;
    nlo = 0 ;
    atx = 0. ;
    /* count bound and free components of x */
    for (j = 0; j < n; j++)
    {
        if      ( x [j] < 1.e-6 )     nlo++ ;
        else if ( x [j] > C - 1.e-6 ) nhi++ ;
        else                          nfree++ ;
        atx += a [j]*x [j] ;
    }
    printf ("constraint violation: %18.10e\n", atx) ;
    printf ("nlo: %i nhi: %i nfree: %i\n\n", nlo, nhi, nfree) ;

    printf ("Timing results:\n") ;
    printf ("    Read matrix:                   %10.4f\n", time1) ;
    printf ("    Construct subspace:            %10.4f\n", Stat.Subspace) ;
    printf ("    Build subspace matrix S:       %10.4f\n", Stat.BuildS) ;
    printf ("    Solve subspace problem in ASL: %10.4f\n", Stat.ASL) ;
    printf ("    Update gradient:               %10.4f\n", Stat.UpdateGrad) ;
    printf ("    Solve problem:                 %10.4f\n", time2) ;
    printf ("    =========================================\n") ;
    printf ("    Total time (read+solve):       %10.4f\n\n", time1+time2) ;
    printf ("Outer iterations:    %6i\n", Stat.outer_its) ;
    printf ("Inner iterations:    %6i\n", Stat.inner_its) ;
    printf ("Columns evaluated:   %6i\n", Stat.ncols) ;
    printf ("Iterations in ASLd:  %6i\n", Stat.nmu) ;
    printf ("Subspace cost evals: %6i\n", Stat.nf) ;
    printf ("Subspace grad evals: %6i\n", Stat.ng) ;
    free (Ax) ;
    free (Ai) ;
    free (Ap) ;
    free (a) ;
    free (x) ;
    i = argc ;
    return (1) ;
}

/* read in a problem (in svmlight format) */
void read_problem
(
    const char *filename, /* name of file containing problem */
    SVMproblem   *PB,
    int         *dim,
    int    *nfeature,
    short   *Ax_is_1
)
{
    short *a, *Ai ;
    int elements, i, j, k, max_index, n, *Ap ;
    float Aj, *Ax ;
    FILE *fp = fopen(filename,"r");

    if(fp == NULL)
    {
        fprintf(stderr,"can't open input file %s\n",filename);
        exit(1);
    }

    n = 0 ;
    elements = 0 ;
    while(1)
    {
        int c = fgetc(fp) ;
        switch(c)
        {
            case '\n':
                ++n ; break ; 
            case ':':
                ++elements ; break;
            case EOF:
                goto out;
            default:
            ;
        }
    }
out:
    rewind(fp);

    PB->a = a = (short *) malloc (n*sizeof (short));
    PB->Ax = Ax = (float  *) malloc (elements*sizeof (float)) ;
    PB->Ai = Ai = (short *) malloc (elements*sizeof (short)) ;
    PB->Ap = Ap = (int *) malloc ((n+1)*sizeof (int)) ;
    *dim = n ;

    Ap [0] = 0 ;
    max_index = 0 ;
    *Ax_is_1 = TRUE ;
    j = 0 ;
    for(i = 0; i < n; i++)
    {
        fscanf(fp,"%d", &k);
        a [i] = k ;

        while(1)
        {
            int c;
            do {
                c = getc(fp) ;
                if ( c=='\n' ) goto out2 ;
            } while (isspace(c)) ;
            ungetc(c,fp);
            if ( fscanf(fp,"%d:%f", &k, &Aj) < 2 )
            {
              fprintf(stderr,"Wrong input format at line %d\n", i+1) ;
              exit(1) ;
            }
            Ai [j] = k-1 ;
            if ( Aj != 1 ) *Ax_is_1 = FALSE ;
            Ax [j] = Aj ;
            j++ ;
        }
out2:
        if ( (j >= 1) && k > max_index ) max_index = k ;
        Ap [i+1] = j ;
    }
    *nfeature = max_index ;
    fclose(fp);
}

static double Timer ( void )
{
    struct rusage ru ;
    double user, tsys ;

    (void) getrusage (RUSAGE_SELF, &ru) ;

    user =
    ru.ru_utime.tv_sec                  /* user time (seconds) */
    + 1e-6 * ru.ru_utime.tv_usec ;      /* user time (microseconds) */

    tsys =
    ru.ru_stime.tv_sec                  /* tsys time (seconds) */
    + 1e-6 * ru.ru_stime.tv_usec ;      /* tsys time (microseconds) */

    return (user + tsys) ;
}
